import uuid
from datetime import datetime
from typing import Optional, Tuple

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.api.v1.endpoints.messaging_endpoints import user_email_invite_status
from fides.api.common_exceptions import AuthorizationError
from fides.api.models.client import ClientDetail
from fides.api.models.fides_user import FidesUser
from fides.api.models.fides_user_invite import FidesUserInvite
from fides.api.schemas.messaging.messaging import (
    MessagingActionType,
    UserInviteBodyParams,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.messaging.message_dispatch_service import dispatch_message
from fides.config import FidesConfig
from fides.config.config_proxy import ConfigProxy


class UserService:
    def __init__(self, db: Session, config: FidesConfig, config_proxy: ConfigProxy):
        self.db = db
        self.config = config
        self.config_proxy = config_proxy

    def invite_user(self, user: FidesUser) -> None:
        """
        Generates a user invite and sends the invite code to the user via email.

        This is a no-op if email messaging isn't configured.
        """

        # invite user via email if email messaging is enabled and the Admin UI URL is defined
        if user_email_invite_status(db=self.db, config_proxy=self.config_proxy).enabled:
            invite_code = str(uuid.uuid4())
            FidesUserInvite.create(
                db=self.db, data={"username": user.username, "invite_code": invite_code}
            )
            user.update(self.db, data={"disabled": True})
            # TODO: refactor to use MessagingService
            dispatch_message(
                self.db,
                action_type=MessagingActionType.USER_INVITE,
                to_identity=Identity(email=user.email_address),
                service_type=self.config_proxy.notifications.notification_service_type,
                message_body_params=UserInviteBodyParams(
                    username=user.username, invite_code=invite_code
                ),
            )
        else:
            logger.debug(
                "Skipping invitation email, an email messaging provider is not enabled",
            )

    def perform_login(
        self,
        client_id_byte_length: int,
        client_secret_byte_length: int,
        user: FidesUser,
        skip_save: Optional[bool] = False,
    ) -> ClientDetail:
        """Performs a login by updating the FidesUser instance and creating and returning
        an associated ClientDetail.

        If the username or password was bad, skip_save should be True. We still run through
        parallel operations to keep the timing of operations similar, but should skip
        saving to the database.
        """

        client = user.client
        if not client:
            logger.info("Creating client for login")
            client, _ = ClientDetail.create_client_and_secret(
                self.db,
                client_id_byte_length,
                client_secret_byte_length,
                scopes=[],  # type: ignore
                roles=user.permissions.roles,  # type: ignore
                systems=user.system_ids,  # type: ignore
                user_id=user.id,
                in_memory=skip_save,  # If login flow has already errored, don't persist this to the database
            )
        else:
            # Refresh the client just in case - for example, scopes and roles were added via the db directly.
            client.roles = user.permissions.roles  # type: ignore
            client.systems = user.system_ids  # type: ignore
            if not skip_save:
                client.save(self.db)

        if user.permissions and (not user.permissions.roles and not user.systems):  # type: ignore
            logger.warning("User {} needs roles or systems to login.", user.id)
            raise AuthorizationError(detail="Not Authorized for this action")

        if not skip_save:
            user.last_login_at = datetime.utcnow()
            user.save(self.db)

        return client

    def accept_invite(
        self, user: FidesUser, new_password: str
    ) -> Tuple[FidesUser, str]:
        """
        Updates the user password and enables the user. Also removes the user invite from the database.
        Returns a tuple of the updated user and their access code.
        """

        # update password and enable
        user.update_password(db=self.db, new_password=new_password)
        user.update(
            self.db,
            data={"disabled": False, "disabled_reason": None},
        )
        self.db.refresh(user)

        # delete invite
        if user.username:
            invite = FidesUserInvite.get_by(
                db=self.db, field="username", value=user.username
            )
            if invite:
                invite.delete(self.db)
        else:
            logger.warning("Username is missing, skipping invite deletion.")

        client = self.perform_login(
            self.config.security.oauth_client_id_length_bytes,
            self.config.security.oauth_client_secret_length_bytes,
            user,
        )

        logger.info("Creating login access token")
        access_code = client.create_access_code_jwe(
            self.config.security.app_encryption_key
        )

        return user, access_code



"""This module handles database credentials for the application database.."""

# pylint: disable=C0115,C0116, E0213

from copy import deepcopy
from typing import Dict, Optional, cast
from urllib.parse import quote, quote_plus, urlencode

from pydantic import Field, PostgresDsn, ValidationInfo, field_validator
from pydantic_settings import SettingsConfigDict

from fides.config.utils import get_test_mode

from .fides_settings import FidesSettings, port_integer_converter

ENV_PREFIX = "FIDES__DATABASE__"


class DatabaseSettings(FidesSettings):
    """Configuration settings for the application database."""

    automigrate: bool = Field(
        default=True,
        description="Automatically runs migrations on webserver startup. If set to `false`, will require the user to run migrations manually via the CLI or API. WARNING: Must be set to `true` for first-time startup.",
    )
    api_engine_pool_size: int = Field(
        default=50,
        description="Number of concurrent database connections Fides will use for API requests. Note that the pool begins with no connections, but as they are requested the connections are maintained and reused up to this limit.",
    )
    api_engine_max_overflow: int = Field(
        default=50,
        description="Number of additional 'overflow' concurrent database connections Fides will use for API requests if the pool reaches the limit. These overflow connections are discarded afterwards and not maintained.",
    )
    api_engine_keepalives_idle: int = Field(
        default=30,
        description="Number of seconds of inactivity before the client sends a TCP keepalive packet to verify the database connection is still alive.",
    )
    api_engine_keepalives_interval: int = Field(
        default=10,
        description="Number of seconds between TCP keepalive retries if the initial keepalive packet receives no response. These are client-side retries.",
    )
    api_engine_keepalives_count: int = Field(
        default=5,
        description="Maximum number of TCP keepalive retries before the client considers the connection dead and closes it.",
    )
    db: str = Field(
        default="default_db", description="The name of the application database."
    )
    load_samples: bool = Field(
        default=False,
        description=(
            "When set to True, initializes the database with sample data for testing (Systems, Datasets, Connectors, etc.) "
            "Used by 'fides deploy' to configure the sample project."
        ),
    )
    password: str = Field(
        default="defaultpassword",
        description="The password with which to login to the application database.",
    )
    port: str = Field(
        default="5432",
        description="The port at which the application database will be accessible.",
    )
    server: str = Field(
        default="default-db",
        description="The hostname of the application database server.",
    )
    readonly_server: Optional[str] = Field(
        default=None,
        description="The hostname of the application read database server.",
    )

    # TODO (LJ-663): add optional readonly_user
    # TODO (LJ-663): add optional readonly_password
    # TODO (LJ-663): add optional readonly_port
    # TODO (LJ-663): add optional readonly_params
    # TODO (LJ-663): add optional readonly_db

    task_engine_pool_size: int = Field(
        default=50,
        description="Number of concurrent database connections Fides will use for executing privacy request tasks, either locally or on each worker. Note that the pool begins with no connections, but as they are requested the connections are maintained and reused up to this limit.",
    )
    task_engine_max_overflow: int = Field(
        default=50,
        description="Number of additional 'overflow' concurrent database connections Fides will use for executing privacy request tasks, either locally or on each worker, if the pool reaches the limit. These overflow connections are discarded afterwards and not maintained.",
    )
    task_engine_keepalives_idle: int = Field(
        default=30,
        description="Number of seconds of inactivity before the client sends a TCP keepalive packet to verify the database connection is still alive.",
    )
    task_engine_keepalives_interval: int = Field(
        default=10,
        description="Number of seconds between TCP keepalive retries if the initial keepalive packet receives no response. These are client-side retries.",
    )
    task_engine_keepalives_count: int = Field(
        default=5,
        description="Maximum number of TCP keepalive retries before the client considers the connection dead and closes it.",
    )
    test_db: str = Field(
        default="default_test_db",
        description="Used instead of the 'db' value when the FIDES_TEST_MODE environment variable is set to True. Avoids overwriting production data.",
        exclude=True,
    )
    user: str = Field(
        default="defaultuser",
        description="The database user with which to login to the application database.",
    )
    params: Dict = Field(
        default={},  # Can't use the default_factory since it breaks docs generation
        description="Additional connection parameters used when connecting to the application database.",
    )

    # These must be at the end because they require other values to construct
    sqlalchemy_database_uri: str = Field(
        default="",
        description="Programmatically created connection string for the application database.",
        exclude=True,
    )
    sqlalchemy_readonly_database_uri: Optional[str] = Field(
        default=None,
        description="Programmatically created connection string for the read-only application database.",
        exclude=True,
    )
    sqlalchemy_test_database_uri: str = Field(
        default="",
        description="Programmatically created connection string for the test database.",
        exclude=True,
    )
    async_database_uri: str = Field(
        default="",
        description="Programmatically created asynchronous connection string for the configured database (either application or test).",
        exclude=True,
    )
    sync_database_uri: str = Field(
        default="",
        description="Programmatically created synchronous connection string for the configured database (either application or test).",
        exclude=True,
    )

    @field_validator("password", mode="before")
    @classmethod
    def escape_password(cls, value: Optional[str]) -> Optional[str]:
        """Escape password"""
        if value and isinstance(value, str):
            return quote_plus(value)
        return value

    @field_validator("sync_database_uri", mode="before")
    @classmethod
    def assemble_sync_database_uri(
        cls, value: Optional[str], info: ValidationInfo
    ) -> str:
        """Join DB connection credentials into a connection string"""
        if isinstance(value, str) and value:
            return value

        port: int = port_integer_converter(info)
        db_name = info.data.get("test_db") if get_test_mode() else info.data.get("db")
        return str(
            PostgresDsn.build(  # pylint: disable=no-member
                scheme="postgresql+psycopg2",
                username=info.data.get("user"),
                password=info.data.get("password"),
                host=info.data.get("server"),
                port=port,
                path=f"{db_name or ''}",
                query=(
                    urlencode(
                        cast(Dict, info.data.get("params")), quote_via=quote, safe="/"
                    )
                    if info.data.get("params")
                    else None
                ),
            )
        )

    @field_validator("async_database_uri", mode="before")
    @classmethod
    def assemble_async_database_uri(
        cls, value: Optional[str], info: ValidationInfo
    ) -> str:
        """Join DB connection credentials into an async connection string."""
        if isinstance(value, str) and value:
            return value

        db_name = info.data.get("test_db") if get_test_mode() else info.data.get("db")

        # Workaround https://github.com/MagicStack/asyncpg/issues/737
        # Required due to the unique way in which Asyncpg handles SSL
        params = cast(Dict, deepcopy(info.data.get("params")))
        if "sslmode" in params:
            params["ssl"] = params.pop("sslmode")
        # This must be constructed in fides.api.db.session as part of the ssl context
        # ref: https://github.com/sqlalchemy/sqlalchemy/discussions/5975
        params.pop("sslrootcert", None)
        # End workaround

        port: int = port_integer_converter(info)
        return str(
            PostgresDsn.build(  # pylint: disable=no-member
                scheme="postgresql+asyncpg",
                username=info.data.get("user"),
                password=info.data.get("password"),
                host=info.data.get("server"),
                port=port,
                path=f"{db_name or ''}",
                query=urlencode(params, quote_via=quote, safe="/") if params else None,
            )
        )

    @field_validator("sqlalchemy_database_uri", mode="before")
    @classmethod
    def assemble_db_connection(cls, v: Optional[str], info: ValidationInfo) -> str:
        """Join DB connection credentials into a synchronous connection string."""
        if isinstance(v, str) and v:
            return v

        port: int = port_integer_converter(info)
        return str(
            PostgresDsn.build(  # pylint: disable=no-member
                scheme="postgresql",
                username=info.data.get("user"),
                password=info.data.get("password"),
                host=info.data.get("server"),
                port=port,
                path=f"{info.data.get('db') or ''}",
                query=(
                    urlencode(
                        cast(Dict, info.data.get("params")), quote_via=quote, safe="/"
                    )
                    if info.data.get("params")
                    else None
                ),
            )
        )

    @field_validator("sqlalchemy_readonly_database_uri", mode="before")
    @classmethod
    def assemble_readonly_db_connection(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Join DB connection credentials into a synchronous connection string."""
        if isinstance(v, str) and v:
            return v
        if not info.data.get("readonly_server"):
            return None
        port: int = port_integer_converter(info)
        return str(
            # TODO: support optional readonly params for user, password, etc.
            PostgresDsn.build(  # pylint: disable=no-member
                scheme="postgresql",
                username=info.data.get("user"),
                password=info.data.get("password"),
                host=info.data.get("readonly_server"),
                port=port,
                path=f"{info.data.get('db') or ''}",
                query=(
                    urlencode(
                        cast(Dict, info.data.get("params")), quote_via=quote, safe="/"
                    )
                    if info.data.get("params")
                    else None
                ),
            )
        )

    @field_validator("sqlalchemy_test_database_uri", mode="before")
    @classmethod
    def assemble_test_db_connection(cls, v: Optional[str], info: ValidationInfo) -> str:
        """Join DB connection credentials into a connection string"""
        if isinstance(v, str) and v:
            return v

        port: int = port_integer_converter(info)
        return str(
            PostgresDsn.build(  # pylint: disable=no-member
                scheme="postgresql",
                username=info.data.get("user"),
                password=info.data.get("password"),
                host=info.data.get("server"),
                port=port,
                path=f"{info.data.get('test_db') or ''}",
                query=(
                    urlencode(
                        cast(Dict, info.data.get("params")), quote_via=quote, safe="/"
                    )
                    if info.data.get("params")
                    else None
                ),
            )
        )

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX, coerce_numbers_to_str=True)



"""Add connection key to execution log

Revision ID: 3caf11127442
Revises: 1f61c765cd1c
Create Date: 2022-12-21 13:36:17.859959

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3caf11127442"
down_revision = "1f61c765cd1c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "executionlog", sa.Column("connection_key", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_executionlog_connection_key"),
        "executionlog",
        ["connection_key"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_executionlog_connection_key"), table_name="executionlog")
    op.drop_column("executionlog", "connection_key")
    # ### end Alembic commands ###



from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_base_aws import (
    BaseAWSSchema,
)


class S3Schema(BaseAWSSchema):
    """Schema to validate the secrets needed to connect to Amazon S3"""


class S3DocsSchema(S3Schema, NoValidationSchema):
    """S3 Secrets Schema for API Docs"""



from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_base_rds import (
    BaseRDSSchema,
)


class RDSMySQLSchema(BaseRDSSchema):
    """
    Schema to validate the secrets needed to connect to a RDS MySQL Database
    """


class RDSMySQLDocsSchema(RDSMySQLSchema, NoValidationSchema):
    """RDS MySQL Secrets Schema for API Docs"""



from pydantic import EmailStr

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    EmailSchema,
)


class AttentiveSchema(EmailSchema):
    third_party_vendor_name: str = "Attentive Email"
    recipient_email_address: EmailStr = "privacy@attentive.com"


class AttentiveDocsSchema(AttentiveSchema, NoValidationSchema):
    """AttentiveDocsSchema Secrets Schema for API Docs"""



import secrets
from functools import cache

from loguru import logger

from fides.api.api.deps import get_autoclose_db_session
from fides.api.models.identity_salt import IdentitySalt


@cache
def get_identity_salt() -> str:
    """
    Checks the database for the identity salt value and creates one if it doesn't exist.
    This function is cached to avoid repeated calls to the database for the same value.
    """

    with get_autoclose_db_session() as db:
        existing_salt = db.query(IdentitySalt).first()
        if existing_salt is None:
            new_salt = IdentitySalt.create(
                db, data={"encrypted_value": {"value": secrets.token_hex(32)}}
            )
            logger.info("Created new identity salt.")
            return new_salt.encrypted_value.get("value")

        logger.info("Caching existing identity salt.")
        return existing_salt.encrypted_value.get("value")



"""update privacy notices for gpp

Revision ID: 956d21f13def
Revises: 61a922702f4c
Create Date: 2024-01-10 02:14:24.802051

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "956d21f13def"
down_revision = "61a922702f4c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacynotice",
        sa.Column(
            "gpp_field_mapping", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column("framework", sa.String(), nullable=True),
    )
    op.alter_column("privacynotice", "data_uses", server_default="{}")

    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "gpp_field_mapping", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column("framework", sa.String(), nullable=True),
    )
    op.alter_column("privacynoticehistory", "data_uses", server_default="{}")

    op.add_column(
        "privacynoticetemplate",
        sa.Column(
            "gpp_field_mapping", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column("framework", sa.String(), nullable=True),
    )
    op.alter_column("privacynoticetemplate", "data_uses", server_default="{}")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacynoticetemplate", "gpp_field_mapping")
    op.drop_column("privacynoticetemplate", "framework")
    op.drop_column("privacynoticehistory", "gpp_field_mapping")
    op.drop_column("privacynoticehistory", "framework")
    op.drop_column("privacynotice", "gpp_field_mapping")
    op.drop_column("privacynotice", "framework")
    # ### end Alembic commands ###



from pydantic import EmailStr

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.connection_configuration.connection_secrets_email import (
    AdvancedSettingsWithExtendedIdentityTypes,
    ExtendedEmailSchema,
    ExtendedIdentityTypes,
)

SOVRN_REQUIRED_IDENTITY: str = "ljt_readerID"


class SovrnSchema(ExtendedEmailSchema):
    """Schema to validate the secrets needed for the SovrnConnector

    Overrides the ExtendedEmailSchema to set the third_party_vendor_name
    and recipient_email_address.

    Also hardcodes the cookie_id for now.
    """

    third_party_vendor_name: str = "Sovrn"
    recipient_email_address: EmailStr = "privacy@sovrn.com"
    advanced_settings: AdvancedSettingsWithExtendedIdentityTypes = (
        AdvancedSettingsWithExtendedIdentityTypes(
            identity_types=ExtendedIdentityTypes(
                email=False, phone_number=False, cookie_ids=[SOVRN_REQUIRED_IDENTITY]
            )
        )
    )


class SovrnDocsSchema(SovrnSchema, NoValidationSchema):
    """SovrnDocsSchema Secrets Schema for API Docs"""



"""add data source params to monitor config

Revision ID: 52a5f1a957bc
Revises: efddde14da21
Create Date: 2024-06-03 20:19:12.481577

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "52a5f1a957bc"
down_revision = "efddde14da21"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "monitorconfig",
        sa.Column(
            "datasource_params", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("monitorconfig", "datasource_params")
    # ### end Alembic commands ###



"""Module that adds functionality for generating or scanning systems."""

import asyncio
from collections import defaultdict
from typing import Dict, List, Optional

from fideslang import manifests
from fideslang.models import Organization, System
from fideslang.validation import AnyHttpUrlString

from fides.common.utils import echo_green, echo_red, handle_cli_response
from fides.connectors.models import AWSConfig, OktaConfig
from fides.core import api
from fides.core.api_helpers import get_server_resource, get_server_resources
from fides.core.filters import filter_aws_systems
from fides.core.parse import parse


def generate_redshift_systems(
    organization_key: str, aws_config: Optional[AWSConfig]
) -> List[System]:
    """
    Fetches Redshift clusters from AWS and returns the transformed System representations.
    """
    import fides.connectors.aws as aws_connector

    client = aws_connector.get_aws_client(service="redshift", aws_config=aws_config)
    describe_clusters = aws_connector.describe_redshift_clusters(client=client)
    redshift_systems = aws_connector.create_redshift_systems(
        describe_clusters=describe_clusters, organization_key=organization_key
    )
    return redshift_systems


def generate_rds_systems(
    organization_key: str, aws_config: Optional[AWSConfig]
) -> List[System]:
    """
    Fetches RDS clusters and instances from AWS and returns the transformed System representations.
    """
    import fides.connectors.aws as aws_connector

    client = aws_connector.get_aws_client(service="rds", aws_config=aws_config)
    describe_clusters = aws_connector.describe_rds_clusters(client=client)
    describe_instances = aws_connector.describe_rds_instances(client=client)
    rds_systems = aws_connector.create_rds_systems(
        describe_clusters=describe_clusters,
        describe_instances=describe_instances,
        organization_key=organization_key,
    )
    return rds_systems


def generate_resource_tagging_systems(
    organization_key: str, aws_config: Optional[AWSConfig]
) -> List[System]:
    """
    Fetches AWS Resources from the resource tagging api and returns the transformed System representations.
    """
    import fides.connectors.aws as aws_connector

    client = aws_connector.get_aws_client(
        service="resourcegroupstaggingapi", aws_config=aws_config
    )
    resources = aws_connector.get_tagging_resources(client=client)
    resource_tagging_systems = aws_connector.create_resource_tagging_systems(
        resources=resources, organization_key=organization_key
    )
    return resource_tagging_systems


def get_organization(
    organization_key: str,
    manifest_organizations: List[Organization],
    url: AnyHttpUrlString,
    headers: Dict[str, str],
) -> Optional[Organization]:
    """
    Attempts to find a given organization by key. Prioritizing a manifest organization
    list and then the server. Raises an error if it cannot find the organization.
    """
    taxonomy_organization = next(
        (org for org in manifest_organizations if org.fides_key == organization_key),
        None,
    )

    if taxonomy_organization:
        return taxonomy_organization

    server_organization = get_server_resource(
        url=str(url),
        resource_type="organization",
        resource_key=organization_key,
        headers=headers,
    )

    if not server_organization:
        echo_red(
            "Could not find organization ({}) to execute command".format(
                organization_key
            )
        )
        raise SystemExit(1)

    parsed_organization = Organization.model_validate(server_organization)
    assert isinstance(parsed_organization, Organization)
    return parsed_organization


def generate_aws_systems(
    organization: Organization, aws_config: Optional[AWSConfig]
) -> List[System]:
    """
    Calls each generate system function for aws resources

    Returns a list of systems with any filters applied
    """
    generate_system_functions = [
        generate_redshift_systems,
        generate_rds_systems,
        generate_resource_tagging_systems,
    ]

    aws_systems = [
        found_system
        for generate_function in generate_system_functions
        for found_system in generate_function(organization.fides_key, aws_config)
    ]

    filtered_aws_systems = filter_aws_systems(
        systems=aws_systems, organization=organization
    )
    return filtered_aws_systems


def write_system_manifest(
    file_name: str, include_null: bool, systems: List[System]
) -> None:
    """
    Given a list of systems, writes a manifest file with the given systems.

    An optional flag can be passed to include null fields.
    """
    manifests.write_manifest(
        file_name,
        [i.model_dump(exclude_none=not include_null) for i in systems],
        "system",
    )
    echo_green(f"Generated system manifest written to {file_name}")


def generate_system_aws(
    file_name: str,
    include_null: bool,
    organization_key: str,
    aws_config: Optional[AWSConfig],
    url: AnyHttpUrlString,
    headers: Dict[str, str],
) -> str:
    """
    Connect to an aws account by leveraging a valid boto3 environment variable
    configuration and extract tracked resource to write a System manifest with.
    """

    organization = get_organization(
        organization_key=organization_key,
        manifest_organizations=[],
        url=url,
        headers=headers,
    )
    assert organization
    aws_systems = generate_aws_systems(organization, aws_config=aws_config)

    write_system_manifest(
        file_name=file_name, include_null=include_null, systems=aws_systems
    )
    return file_name


async def generate_okta_systems(
    organization: Organization, okta_config: Optional[OktaConfig]
) -> List[System]:
    """
    Given an okta configuration, calls okta for a list of
    applications and returns the corresponding systems.
    """

    import fides.connectors.okta as okta_connector

    okta_client = okta_connector.get_okta_client(okta_config)
    okta_applications = await okta_connector.list_okta_applications(
        okta_client=okta_client
    )
    okta_systems = okta_connector.create_okta_systems(
        okta_applications=okta_applications, organization_key=organization.fides_key
    )
    return okta_systems


def generate_system_okta(
    organization_key: str,
    okta_config: Optional[OktaConfig],
    file_name: str,
    include_null: bool,
    url: AnyHttpUrlString,
    headers: Dict[str, str],
) -> str:
    """
    Generates a system manifest from existing Okta applications.
    """

    organization = get_organization(
        organization_key=organization_key,
        manifest_organizations=[],
        url=url,
        headers=headers,
    )
    assert organization

    okta_systems = asyncio.run(
        generate_okta_systems(organization=organization, okta_config=okta_config)
    )

    write_system_manifest(
        file_name=file_name, include_null=include_null, systems=okta_systems
    )
    return file_name


def get_all_server_systems(
    url: AnyHttpUrlString, headers: Dict[str, str], exclude_systems: List[System]
) -> List[System]:
    """
    Get a list of all of the Systems that exist on the server. Excludes any systems
    provided in exclude_systems
    """
    ls_response = handle_cli_response(
        api.ls(url=str(url), resource_type="system", headers=headers), verbose=False
    )
    exclude_system_keys = [system.fides_key for system in exclude_systems]
    system_keys = [
        resource["fides_key"]
        for resource in ls_response.json()
        if resource["fides_key"] not in exclude_system_keys
    ]

    # The validation here is required to guarantee the return type
    system_list = [
        System.validate(x)
        for x in get_server_resources(
            url=str(url),
            resource_type="system",
            headers=headers,
            existing_keys=system_keys,
        )
    ]

    return system_list


def get_scan_resource_label(system_type: str) -> str:
    """
    Given a system type, returns the label to use in scan output
    """
    resource_label_map = {
        "redshift_cluster": "Redshift Cluster",
        "rds_instance": "RDS Instance",
        "rds_cluster": "RDS Cluster",
        "okta_application": "Okta Application",
    }
    resource_label = resource_label_map.get(system_type, "Unknown")
    return resource_label


def get_system_resource_ids(systems: List[System]) -> List[str]:
    """
    Given a list of systems, build a list of resource ids
    """
    system_resource_ids = []
    for system in systems:
        resource_id = get_system_resource_id(system=system)
        if resource_id:
            system_resource_ids.append(resource_id)
    return system_resource_ids


def get_system_resource_id(system: System) -> Optional[str]:
    """
    Given a resource system, returns the resource id. Returns None
    if a resource_id is not found in metadata.
    """
    system_resource_id = (
        system.fidesctl_meta.resource_id
        if system.fidesctl_meta and system.fidesctl_meta.resource_id
        else None
    )
    return system_resource_id


def find_missing_systems(
    source_systems: List[System],
    existing_systems: List[System],
) -> List[System]:
    """
    Given a list of source and existing systems, returns a list
    of systems which are not found in the existing list.
    """
    existing_resource_ids = get_system_resource_ids(systems=existing_systems)
    missing_systems = [
        source_system
        for source_system in source_systems
        if get_system_resource_id(system=source_system) not in existing_resource_ids
    ]
    return missing_systems


def print_system_resource_scan_result(
    source_systems: List[System],
    missing_systems: List[System],
    coverage_threshold: int,
) -> None:
    """
    Prints missing systems and raises an exception if coverage
    is lower than provided threshold.
    """
    output: str = "Successfully scanned the following systems:\n"
    for source_system in source_systems:
        output += "\t{}(id={})\n".format(
            source_system.name, get_system_resource_id(system=source_system)
        )
    echo_green(output)

    if missing_systems:
        missing_systems_output = (
            "The following system were not found in existing manifest\n"
        )

        # first group the systems by type to print them grouped
        missing_systems_by_type = defaultdict(list)
        for missing_system in missing_systems:
            missing_systems_by_type[missing_system.system_type].append(missing_system)

        for typed_missing_systems in missing_systems_by_type.values():
            resource_label = get_scan_resource_label(
                system_type=typed_missing_systems[0].system_type
            )
            missing_systems_output += f"Missing {resource_label}:\n"

            for typed_missing_system in typed_missing_systems:
                missing_systems_output += "\t{}(id={})\n".format(
                    typed_missing_system.name,
                    get_system_resource_id(system=typed_missing_system),
                )
        echo_red(missing_systems_output)

    coverage_percent = int(
        ((len(source_systems) - len(missing_systems)) / len(source_systems)) * 100
    )
    annotation_output = "Resource coverage: {}%".format(coverage_percent)
    if coverage_percent < coverage_threshold:
        echo_red(annotation_output)
        raise SystemExit(1)
    echo_green(annotation_output)


def scan_system_aws(
    manifest_dir: str,
    organization_key: str,
    aws_config: Optional[AWSConfig],
    coverage_threshold: int,
    url: AnyHttpUrlString,
    headers: Dict[str, str],
) -> None:
    """
    Connect to an aws account by leveraging a valid boto3 environment variable
    configuration and compares tracked resources to existing systems.
    """

    manifest_taxonomy = parse(manifest_dir) if manifest_dir else None
    manifest_systems = (
        manifest_taxonomy.system
        if manifest_taxonomy and manifest_taxonomy.system
        else []
    )
    server_systems = get_all_server_systems(
        url=url, headers=headers, exclude_systems=manifest_systems
    )
    existing_systems = manifest_systems + server_systems

    organization = get_organization(
        organization_key=organization_key,
        manifest_organizations=(
            manifest_taxonomy.organization if manifest_taxonomy else []
        ),
        url=url,
        headers=headers,
    )
    assert organization

    aws_systems = generate_aws_systems(organization=organization, aws_config=aws_config)

    if len(aws_systems) < 1:
        echo_red("AWS did not return any resources to scan systems")
        raise SystemExit(1)

    missing_systems = find_missing_systems(
        source_systems=aws_systems,
        existing_systems=existing_systems,
    )
    print_system_resource_scan_result(
        source_systems=aws_systems,
        missing_systems=missing_systems,
        coverage_threshold=coverage_threshold,
    )


def scan_system_okta(
    manifest_dir: str,
    organization_key: str,
    okta_config: Optional[OktaConfig],
    coverage_threshold: int,
    url: AnyHttpUrlString,
    headers: Dict[str, str],
) -> None:
    """
    Fetches Okta applications and compares them against existing
    systems in the server and manifest supplied.
    """

    manifest_taxonomy = parse(manifest_dir) if manifest_dir else None
    manifest_systems = (
        manifest_taxonomy.system
        if manifest_taxonomy and manifest_taxonomy.system
        else []
    )
    server_systems = get_all_server_systems(
        url=url, headers=headers, exclude_systems=manifest_systems
    )

    organization = get_organization(
        organization_key=organization_key,
        manifest_organizations=(
            manifest_taxonomy.organization if manifest_taxonomy else []
        ),
        url=url,
        headers=headers,
    )
    assert organization

    okta_systems = asyncio.run(
        generate_okta_systems(organization=organization, okta_config=okta_config)
    )

    if len(okta_systems) < 1:
        echo_red("Okta did not return any applications to scan systems")
        raise SystemExit(1)

    missing_systems = find_missing_systems(
        source_systems=okta_systems,
        existing_systems=server_systems + manifest_systems,
    )
    print_system_resource_scan_result(
        source_systems=okta_systems,
        missing_systems=missing_systems,
        coverage_threshold=coverage_threshold,
    )



"""privacy_preferences_v2_data

Revision ID: 5a8cee9c014c
Revises: f9b28f36b53e
Create Date: 2023-12-10 20:41:16.804029

"""

import json
from enum import Enum
from typing import Dict, List, Optional, Set

import networkx as nx
import pandas as pd
import sqlalchemy_utils
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
from pandas import DataFrame, Series
from sqlalchemy import String, text
from sqlalchemy.engine import Connection
from sqlalchemy_utils.types.encrypted.encrypted_type import AesGcmEngine

from fides.api.db.base_class import JSONTypeOverride
from fides.config import CONFIG

revision = "5a8cee9c014c"
down_revision = "f9b28f36b53e"
branch_labels = None
depends_on = None


def upgrade():
    """Data migration for overhaul around how Privacy Preferences are saved.
    Historical preferences saved and notices served are migrated in place while
    current preferences and last notices served are migrated to other tables.
    """

    bind = op.get_bind()

    logger.info("Migrating PrivacyPreferenceHistory.")

    # Deleting preferences saved against TCF as this is not considered live
    bind.execute(text(TCF_PREFERENCES_DELETE_QUERY))

    # Migrate over Notice Name, Key, and Mechanism from Privacy Notice
    bind.execute(text(PRIVACY_PREFERENCE_HISTORY_UPDATE_QUERY))

    logger.info("Migrating ServedNoticeHistory.")

    # Deleting TCF attributes served as this is not yet considered live.
    bind.execute(text(TCF_SERVED_DELETE_QUERY))

    # Migrating over notice name, key, mechanism and served notice history id.
    # Going forward a served notice history id will be generated for the entire request before
    # it is returned to the frontend, but for this migration records, use the id itself
    bind.execute(
        text(SERVED_NOTICE_HISTORY_UPDATE_QUERY),
    )

    logger.info("Migrating CurrentPrivacyPreference to CurrentPrivacyPreferenceV2")
    # Collapses preferences from to a single user into the same record, prioritizing more recent identifiers
    # and preferences
    migrate_current_records(
        bind,
        CURRENT_PRIVACY_PREFERENCE_BASE_QUERY,
        CurrentMigrationType.preferences,
    )

    logger.info("Migrating LastServedNotice to LastServedNoticeV2")
    # Collapses preferences from to a single user into the same record, prioritizing more recent identifiers
    migrate_current_records(
        bind, LAST_SERVED_NOTICE_BASE_QUERY, CurrentMigrationType.served
    )


def downgrade():
    logger.info("Downgrade: Reversing PrivacyPreferenceHistory Notice Updates.")

    bind = op.get_bind()
    bind.execute(text(PRIVACY_PREFERENCE_HISTORY_UPDATE_DOWNREV_QUERY))

    logger.info("Downgrade: Reversing ServedNoticeHistory Notice Updates.")
    bind.execute(text(SERVED_NOTICE_HISTORY_UPDATE_DOWNREV_QUERY))

    logger.info("Downgrade: Delete data from CurrentPrivacyPreferenceV2.")
    bind.execute(text("""DELETE FROM currentprivacypreferencev2"""))

    logger.info("Downgrade: Delete data from LastServedNoticeV2.")
    bind.execute(text("""DELETE FROM lastservednoticev2"""))


PRIVACY_PREFERENCE_HISTORY_UPDATE_QUERY = """
    UPDATE privacypreferencehistory
    SET 
        notice_name = privacynoticehistory.name,
        notice_key = privacynoticehistory.notice_key,
        notice_mechanism = privacynoticehistory.consent_mechanism
    FROM privacynoticehistory
    WHERE privacypreferencehistory.privacy_notice_history_id = privacynoticehistory.id         
"""

PRIVACY_PREFERENCE_HISTORY_UPDATE_DOWNREV_QUERY = """
    UPDATE privacypreferencehistory
    SET 
        notice_name = null,
        notice_key = null,
        notice_mechanism = null;    
"""

TCF_PREFERENCES_DELETE_QUERY = """
    DELETE FROM privacypreferencehistory WHERE privacy_notice_history_id IS NULL;        
"""


SERVED_NOTICE_HISTORY_UPDATE_QUERY = """
    UPDATE servednoticehistory
    SET 
        notice_name = privacynoticehistory.name,
        notice_key = privacynoticehistory.notice_key,
        notice_mechanism = privacynoticehistory.consent_mechanism,
        served_notice_history_id = servednoticehistory.id
    FROM privacynoticehistory
    WHERE servednoticehistory.privacy_notice_history_id = privacynoticehistory.id         
"""

SERVED_NOTICE_HISTORY_UPDATE_DOWNREV_QUERY = """
    UPDATE servednoticehistory
    SET 
        notice_name = null,
        notice_key = null,
        served_notice_history_id = null,
        notice_mechanism = null;    
"""

TCF_SERVED_DELETE_QUERY = """
    DELETE FROM servednoticehistory WHERE privacy_notice_history_id IS NULL;        
"""

CURRENT_PRIVACY_PREFERENCE_BASE_QUERY = """
    SELECT
        currentprivacypreference.id,
        currentprivacypreference.preference,
        currentprivacypreference.privacy_notice_history_id,
        email_details.hashed_value as hashed_email, 
        device_details.hashed_value as hashed_fides_user_device,
        phone_details.hashed_value as hashed_phone_number,
        email_details.encrypted_value as encrypted_email, 
        device_details.encrypted_value as encrypted_device,
        phone_details.encrypted_value as encrypted_phone,
        currentprivacypreference.created_at,
        currentprivacypreference.updated_at
    FROM currentprivacypreference
    LEFT OUTER JOIN providedidentity AS email_details ON email_details.field_name = 'email' AND email_details.id = currentprivacypreference.provided_identity_id
    LEFT OUTER JOIN providedidentity AS phone_details ON phone_details.field_name = 'phone_number' AND phone_details.id = currentprivacypreference.provided_identity_id
    LEFT OUTER JOIN providedidentity AS device_details ON device_details.field_name = 'fides_user_device_id' AND device_details.id = currentprivacypreference.fides_user_device_provided_identity_id
    WHERE privacy_notice_history_id IS NOT NULL
    ORDER BY created_at asc;
"""


LAST_SERVED_NOTICE_BASE_QUERY = """
    SELECT
        lastservednotice.id,
        lastservednotice.privacy_notice_history_id,
        email_details.hashed_value as hashed_email, 
        device_details.hashed_value as hashed_fides_user_device,
        phone_details.hashed_value as hashed_phone_number,
        email_details.encrypted_value as encrypted_email, 
        device_details.encrypted_value as encrypted_device,
        phone_details.encrypted_value as encrypted_phone,
        lastservednotice.created_at,
        lastservednotice.updated_at
    FROM lastservednotice
    LEFT OUTER JOIN providedidentity AS email_details ON email_details.field_name = 'email' AND email_details.id = lastservednotice.provided_identity_id
    LEFT OUTER JOIN providedidentity AS phone_details ON phone_details.field_name = 'phone_number' AND phone_details.id = lastservednotice.provided_identity_id
    LEFT OUTER JOIN providedidentity AS device_details ON device_details.field_name = 'fides_user_device_id' AND device_details.id = lastservednotice.fides_user_device_provided_identity_id
    WHERE privacy_notice_history_id IS NOT NULL
    ORDER BY created_at asc;
"""


class CurrentMigrationType(Enum):
    preferences = "preferences"
    served = "served"


encryptor = sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(
    String,
    CONFIG.security.app_encryption_key,
    AesGcmEngine,
    "pkcs5",
)

decryptor = sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(
    JSONTypeOverride,
    CONFIG.security.app_encryption_key,
    AesGcmEngine,
    "pkcs5",
)


def migrate_current_records(
    bind: Connection, starting_query: str, migration_type: CurrentMigrationType
):
    """Common method to migrate CurrentPrivacyPreference -> CurrentPrivacyPreferenceV2 and
    LastServedNotice -> LastServedNoticeV2.

    We are migrating from tables with unique constraints on two provided identity types (x) preferences types to the new
    tables with unique constraints on email, device id, and phone number.

    Migration involves linking all records in the original table with any shared identifiers across email, phone,
    or device id, and collapsing these records into single rows, retaining the most recently used non-null identifiers
    and recently saved preferences.
    """
    df: DataFrame = pd.read_sql(starting_query, bind)

    if len(df.index) == 0:
        logger.info(f"No {migration_type.value} records to migrate. Skipping.")
        return

    # Drop invalid rows where we have an encrypted val but not a hashed val and vice versa.
    # This would be unexpected, but this would mean our ProvidedIdentity record was not populated correctly.
    df["email_count"] = df[["encrypted_email", "hashed_email"]].count(axis=1)
    df["phone_count"] = df[["encrypted_phone", "hashed_phone_number"]].count(axis=1)
    df["device_count"] = df[["encrypted_device", "hashed_fides_user_device"]].count(
        axis=1
    )
    df = df[df["email_count"] != 1]
    df = df[df["phone_count"] != 1]
    df = df[df["device_count"] != 1]

    # Also drop if there are no identifiers at all - our new table needs at least one
    df = df[df["email_count"] + df["phone_count"] + df["device_count"] >= 2]

    # Create a "paths" column in the dataframe that is a list of non-null identifiers, so
    # we only consider actual values as a match.
    df["paths"] = df[
        ["hashed_email", "hashed_phone_number", "hashed_fides_user_device"]
    ].apply(lambda row: [val for val in row if pd.notna(val)], axis=1)

    network_x_graph: nx.Graph = nx.Graph()
    # Adds every path to the Graph
    df["paths"].apply(lambda path: nx.add_path(network_x_graph, path))

    # This is the magic - linking any common records across hashed_email OR hashed_phone OR hashed_device
    connected_records: List[Set] = list(nx.connected_components(network_x_graph))

    def add_group_id_based_on_link(identity_path: List[str]) -> int:
        """Add a common group id for records that belong to the same connected component"""
        for user_identifier in identity_path:
            for i, linked_nodes in enumerate(connected_records):
                if user_identifier in linked_nodes:
                    return i + 1

    df["group_id"] = df["paths"].apply(add_group_id_based_on_link)

    result_df = (
        _group_preferences_records(df)
        if migration_type == CurrentMigrationType.preferences
        else _group_served_records(df)
    )

    def decrypt_extract_encrypt(
        provided_identity_encrypted_value: Optional[str],
    ) -> Optional[str]:
        """Decrypt the Provided Identity encrypted value, then extract the value from {"value": xxxx} and re-encrypt xxxx"""
        if not provided_identity_encrypted_value:
            return None

        decrypted = (
            decryptor.process_result_value(
                provided_identity_encrypted_value, dialect=""
            )
            or {}
        ).get("value")

        return encryptor.process_bind_param(decrypted, dialect="")

    # Encrypted value is stored differently on ProvidedIdentity than this table.  Decrypt, extract the value,
    # then re-encrypt.
    result_df["email"] = result_df["encrypted_email"].apply(decrypt_extract_encrypt)
    result_df["phone_number"] = result_df["encrypted_phone"].apply(
        decrypt_extract_encrypt
    )
    result_df["fides_user_device"] = result_df["encrypted_device"].apply(
        decrypt_extract_encrypt
    )

    # Remove columns from aggregated data frame that are not needed in CurrentPrivacyPreferenceV2 or
    # LastServedNoticeV2 table before writing new data
    result_df.drop(columns="group_id", inplace=True)
    result_df.drop(columns="encrypted_email", inplace=True)
    result_df.drop(columns="encrypted_phone", inplace=True)
    result_df.drop(columns="encrypted_device", inplace=True)

    if migration_type == CurrentMigrationType.preferences:
        result_df.to_sql(
            "currentprivacypreferencev2", con=bind, if_exists="append", index=False
        )
    else:
        result_df.to_sql(
            "lastservednoticev2", con=bind, if_exists="append", index=False
        )


def _group_preferences_records(df: DataFrame) -> DataFrame:
    """Combine preferences belonging to the same user under our definition.

    Collapse records into rows by group_id, combining identifiers and preferences against privacy notice history ids,
    retaining the most recently saved"""

    # Add a preferences column, combining privacy_notice_history_id and preference
    df["preferences"] = df.apply(
        lambda row: (row["privacy_notice_history_id"], row["preference"]), axis=1
    )

    def combine_preferences(preferences: Series) -> str:
        """Combines the preferences across user records deemed to be linked, prioritizing most recently saved due to
        sort order"""
        prefs: Dict = {}
        for preference in preferences:
            # Records were sorted ascending by date, so last one in wins (most recently saved)
            prefs[preference[0]] = preference[1]

        return json.dumps(
            {
                "preferences": [
                    {
                        "privacy_notice_history_id": notice_history,
                        "preference": preference,
                    }
                    for notice_history, preference in prefs.items()
                ],
                "purpose_consent_preferences": [],
                "purpose_legitimate_interests_preferences": [],
                "special_purpose_preferences": [],
                "feature_preferences": [],
                "special_feature_preferences": [],
                "vendor_consent_preferences": [],
                "vendor_legitimate_interests_preferences": [],
                "system_consent_preferences": [],
                "system_legitimate_interests_preferences": [],
            }
        )

    # Groups by group_id, prioritizing latest non-null records for identifiers, and more recently saved privacy
    # preferences.
    result_df = (
        df.groupby("group_id")
        .agg(
            id=("id", "last"),
            hashed_email=("hashed_email", "last"),
            hashed_phone_number=("hashed_phone_number", "last"),
            hashed_fides_user_device=("hashed_fides_user_device", "last"),
            created_at=("created_at", "last"),
            updated_at=("updated_at", "last"),
            encrypted_email=("encrypted_email", "last"),
            encrypted_phone=("encrypted_phone", "last"),
            encrypted_device=("encrypted_device", "last"),
            preferences=("preferences", combine_preferences),
        )
        .reset_index()
    )
    return result_df


def _group_served_records(df: DataFrame):
    """Collapse records into rows on group_id, combining identifiers privacy notices served"""

    def combine_served(served: Series) -> str:
        """Combines the preferences across user records deemed to be linked, prioritizing most recently saved due to
        sort order"""
        return json.dumps(
            {
                "privacy_notice_history_ids": served.unique().tolist(),
                "tcf_purpose_consents": [],
                "tcf_purpose_legitimate_interests": [],
                "tcf_special_purposes": [],
                "tcf_vendor_consents": [],
                "tcf_vendor_legitimate_interests": [],
                "tcf_features": [],
                "tcf_special_features": [],
                "tcf_system_consents": [],
                "tcf_system_legitimate_interests": [],
            }
        )

    # Groups by group_id, prioritizing latest non-null records for identifiers, and more recently saved privacy
    # preferences.
    result_df = (
        df.groupby("group_id")
        .agg(
            id=("id", "last"),
            hashed_email=("hashed_email", "last"),
            hashed_phone_number=("hashed_phone_number", "last"),
            hashed_fides_user_device=("hashed_fides_user_device", "last"),
            created_at=("created_at", "last"),
            updated_at=("updated_at", "last"),
            encrypted_email=("encrypted_email", "last"),
            encrypted_phone=("encrypted_phone", "last"),
            encrypted_device=("encrypted_device", "last"),
            served=("privacy_notice_history_id", combine_served),
        )
        .reset_index()
    )
    return result_df



from fides.api.service.connectors.postgres_connector import PostgreSQLConnector


class TimescaleConnector(PostgreSQLConnector):
    """TimescaleDB Connector - This currently supports identical functionality to the PostgreSQL connector"""



"""Populate saas type to saas config

Revision ID: fc90277bbcde
Revises: ed1b00ff963d
Create Date: 2022-07-05 19:20:59.384767

"""

import enum
from typing import List

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "fc90277bbcde"
down_revision = "c7cc36820d4b"
branch_labels = None
depends_on = None


class SaaSType(enum.Enum):
    """
    Enum to store saas connection type in Fidesops
    """

    mailchimp = "mailchimp"
    hubspot = "hubspot"
    outreach = "outreach"
    segment = "segment"
    sentry = "sentry"
    stripe = "stripe"
    zendesk = "zendesk"
    custom = "custom"


query = text(
    """select connectionconfig.id, connectionconfig.saas_config from connectionconfig WHERE connectionconfig.connection_type = :connection_type"""
)
update_query = text(
    """update connectionconfig set saas_config = jsonb_set(saas_config, '{type}', :saas_type) where id = :id"""
)


def upgrade():
    connection = op.get_bind()
    saas_options: List[str] = [saas_type.value for saas_type in SaaSType]
    for id, saas_config in connection.execute(query, {"connection_type": "saas"}):
        if not saas_config:
            continue
        fides_key: str = saas_config.get("fides_key", "")
        saas_name: str = saas_config.get("name", "")

        try:
            saas_type: str = next(
                (
                    opt
                    for opt in saas_options
                    if any(
                        [
                            fides_key.lower().startswith(opt),
                            saas_name.lower().startswith(opt),
                        ]
                    )
                ),
                "custom",
            )
            connection.execute(update_query, {"saas_type": f'"{saas_type}"', "id": id})
        except Exception:
            # default to using "custom" if something went wrong
            connection.execute(update_query, {"saas_type": f'"custom"', "id": id})


def downgrade():
    connection = op.get_bind()
    for id, saas_config in connection.execute(query, {"connection_type": "saas"}):
        connection.execute(update_query, {"saas_type": f'"custom"', "id": id})



import json
import os
import zipfile
from collections import defaultdict
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Optional

import jinja2
from jinja2 import Environment, FileSystemLoader

from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.util.storage_util import storage_json_encoder

DSR_DIRECTORY = Path(__file__).parent.resolve()

TEXT_COLOR = "#4A5568"
HEADER_COLOR = "#FAFAFA"
BORDER_COLOR = "#E2E8F0"


# pylint: disable=too-many-instance-attributes
class DsrReportBuilder:
    def __init__(
        self,
        privacy_request: PrivacyRequest,
        dsr_data: Dict[str, Any],
    ):
        """
        Manages populating HTML templates from the given data and adding the generated
        pages to a zip file in a way that the pages can be navigated between.
        """

        # zip file variables
        self.baos = BytesIO()

        # we close this in the finally block of generate()
        # pylint: disable=consider-using-with
        self.out = zipfile.ZipFile(self.baos, "w")

        # Jinja template environment initialization
        def pretty_print(value: str, indent: int = 4) -> str:
            return json.dumps(value, indent=indent, default=storage_json_encoder)

        jinja2.filters.FILTERS["pretty_print"] = pretty_print
        self.template_loader = Environment(
            loader=FileSystemLoader(DSR_DIRECTORY), autoescape=True
        )

        # to pass in custom colors in the future
        self.template_data: Dict[str, Any] = {
            "text_color": TEXT_COLOR,
            "header_color": HEADER_COLOR,
            "border_color": BORDER_COLOR,
        }
        self.main_links: Dict[str, Any] = {}  # used to track the generated pages

        # report data to populate the templates
        self.request_data = _map_privacy_request(privacy_request)
        self.dsr_data = dsr_data

    def _populate_template(
        self,
        template_path: str,
        heading: Optional[str] = None,
        description: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Generates a file from the template and data"""
        report_data = {
            "heading": heading,
            "description": description,
            "data": data,
            "request": self.request_data,
        }
        report_data.update(self.template_data)
        template = self.template_loader.get_template(template_path)
        rendered_template = template.render(report_data)
        return rendered_template

    def _add_file(self, filename: str, contents: str) -> None:
        """Helper to add a file to the zip archive"""
        if filename and contents:
            self.out.writestr(f"{filename}", contents.encode("utf-8"))

    def _add_dataset(self, dataset_name: str, collections: Dict[str, Any]) -> None:
        """
        Generates a page for each collection in the dataset and an index page for the dataset.
        Tracks the generated links to build a root level index after each collection has been processed.
        """
        # track links to collection indexes
        collection_links = {}
        for collection_name, rows in collections.items():
            collection_url = f"{collection_name}/index.html"
            self._add_collection(rows, dataset_name, collection_name)
            collection_links[collection_name] = collection_url

        # generate dataset index page
        self._add_file(
            f"data/{dataset_name}/index.html",
            self._populate_template(
                "templates/dataset_index.html",
                dataset_name,
                None,
                collection_links,
            ),
        )

    def _add_collection(
        self, rows: List[Dict[str, Any]], dataset_name: str, collection_name: str
    ) -> None:
        # track links to detail pages
        detail_links = {}
        for index, item in enumerate(rows, 1):
            detail_url = f"{index}.html"
            self._add_file(
                f"data/{dataset_name}/{collection_name}/{index}.html",
                self._populate_template(
                    "templates/item.html",
                    f"{collection_name} (item #{index})",
                    None,
                    item,
                ),
            )
            detail_links[f"item #{index}"] = detail_url

        # generate detail index page
        self._add_file(
            f"data/{dataset_name}/{collection_name}/index.html",
            self._populate_template(
                "templates/collection_index.html",
                collection_name,
                None,
                detail_links,
            ),
        )

    def generate(self) -> BytesIO:
        """
        Processes the request and DSR data to build zip file containing the DSR report.
        Returns the zip file as an in-memory byte array.
        """
        try:
            # all the css for the pages is in main.css
            self._add_file(
                "data/main.css",
                self._populate_template("templates/main.css"),
            )
            self._add_file(
                "data/back.svg",
                Path(os.path.join(DSR_DIRECTORY, "assets/back.svg")).read_text(
                    encoding="utf-8"
                ),
            )

            # pre-process data to split the dataset:collection keys
            datasets: Dict[str, Any] = defaultdict(lambda: defaultdict(list))
            for key, rows in self.dsr_data.items():
                parts = key.split(":", 1)
                dataset_name, collection_name = (
                    parts if len(parts) > 1 else ("manual", parts[0])
                )
                datasets[dataset_name][collection_name].extend(rows)

            for dataset_name, collections in datasets.items():
                self._add_dataset(dataset_name, collections)
                self.main_links[dataset_name] = f"data/{dataset_name}/index.html"

            # create the main index once all the datasets have been added
            self._add_file(
                "welcome.html",
                self._populate_template(
                    "templates/welcome.html", "DSR Report", None, self.main_links
                ),
            )
        finally:
            # close out zip file in the finally block to always close, even when an exception occurs
            self.out.close()

        # reset the file pointer so the file can be fully read by the caller
        self.baos.seek(0)
        return self.baos


def _map_privacy_request(privacy_request: PrivacyRequest) -> Dict[str, Any]:
    """Creates a map with a subset of values from the privacy request"""
    request_data: Dict[str, Any] = {}
    request_data["id"] = privacy_request.id

    action_type: Optional[ActionType] = privacy_request.policy.get_action_type()
    if action_type:
        request_data["type"] = action_type.value

    request_data["identity"] = {
        key: value
        for key, value in privacy_request.get_persisted_identity()
        .labeled_dict(include_default_labels=True)
        .items()
        if value["value"] is not None
    }

    if privacy_request.requested_at:
        request_data["requested_at"] = privacy_request.requested_at.strftime(
            "%m/%d/%Y %H:%M %Z"
        )
    return request_data



from __future__ import annotations

from datetime import datetime
from enum import Enum
from re import match
from typing import Any, Dict, Iterable, List, Optional, Type

from loguru import logger
from sqlalchemy import ARRAY, Boolean, Column, DateTime, ForeignKey, String, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.future import select
from sqlalchemy.orm import Session, relationship
from sqlalchemy.orm.query import Query

from fides.api.db.base_class import Base, FidesBase
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.sql_models import System  # type: ignore[attr-defined]


class DiffStatus(Enum):
    ADDITION = "addition"
    REMOVAL = "removal"
    CLASSIFICATION_ADDITION = "classification_addition"
    CLASSIFICATION_UPDATE = "classification_update"
    MONITORED = "monitored"
    MUTED = "muted"


class MonitorFrequency(Enum):
    """
    Enum representing monitor frequency. Not used in DB but needed for translating to API schema
    """

    DAILY = "Daily"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    YEARLY = "Yearly"
    NOT_SCHEDULED = "Not scheduled"


# pattern for a string of 4 comma-separated integers,
# used to represent the months of the year that the monitor will run
# on quarterly basis, in cron format
QUARTERLY_MONTH_PATTERN = r"^\d+,\d+,\d+,\d+$"


class MonitorConfig(Base):
    """
    Monitor configuration used for data detection and discovery.

    Each monitor configuration references `ConnectionConfig`, which provide it with underlying
    configuration details used in connecting to the external data store.
    """

    name = Column(String, nullable=True)
    key = Column(String, index=True, unique=True, nullable=False)
    connection_config_id = Column(
        String,
        ForeignKey(ConnectionConfig.id_field_path),
        nullable=False,
        index=True,
    )
    databases = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        server_default="{}",
        default=dict,
    )  # the databases to which the monitor is scoped
    excluded_databases = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        server_default="{}",
        default=dict,
    )  # the databases to which the monitor is not scoped
    monitor_execution_trigger = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # stores the cron-based kwargs for scheduling the monitor execution.
    # see https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html

    classify_params = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # parameters that the monitor will use for classification execution

    datasource_params = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # monitor parameters that are specific per datasource
    # these are held as an untyped JSON dict (in the DB) to stay flexible

    last_monitored = Column(
        DateTime(timezone=True),
        nullable=True,
    )  # when the monitor was last executed

    enabled = Column(
        Boolean,
        default=True,
        nullable=False,
        server_default="t",
    )

    # TODO: many-to-many link to users assigned as data stewards; likely will need a join-table

    connection_config = relationship(ConnectionConfig)

    executions = relationship(
        "MonitorExecution",
        cascade="all, delete-orphan",
        backref="monitor_config",
    )

    @property
    def connection_config_key(self) -> str:
        """Derives the `connection_config_key`"""
        return self.connection_config.key

    @property
    def execution_start_date(self) -> Optional[datetime]:
        """Derives the `execution_start_date`"""
        if (
            not self.monitor_execution_trigger
            or self.monitor_execution_trigger.get("start_date", None) is None
        ):
            return None
        return self.monitor_execution_trigger.get("start_date")

    @property
    def execution_frequency(self) -> Optional[MonitorFrequency]:
        """Derives the `execution_frequency`"""
        if (
            not self.monitor_execution_trigger
            or self.monitor_execution_trigger.get("hour", None) is None
        ):
            return MonitorFrequency.NOT_SCHEDULED
        month_trigger = self.monitor_execution_trigger.get("month", None)
        if month_trigger is not None:
            if isinstance(month_trigger, str) and match(
                QUARTERLY_MONTH_PATTERN, month_trigger
            ):
                return MonitorFrequency.QUARTERLY
            return MonitorFrequency.YEARLY
        if self.monitor_execution_trigger.get("day", None) is not None:
            return MonitorFrequency.MONTHLY
        if self.monitor_execution_trigger.get("day_of_week", None) is not None:
            return MonitorFrequency.WEEKLY
        return MonitorFrequency.DAILY

    def update(self, db: Session, *, data: dict[str, Any]) -> FidesBase:
        """
        Override the base class `update` to validate database include/exclude
        and derive the `execution_trigger` dict field
        """
        MonitorConfig.database_include_exclude_list_is_valid(data)
        MonitorConfig.derive_execution_trigger_dict(data)
        return super().update(db=db, data=data)

    @classmethod
    def database_include_exclude_list_is_valid(cls, data: Dict[str, Any]) -> None:
        """Check that both include and exclude have not both been set"""
        include = data.get("databases", [])
        exclude = data.get("excluded_databases", [])
        if include and exclude:
            raise ValueError(
                "Both `databases` and `excluded_databases` cannot be set at the same time."
            )

    @classmethod
    def create(
        cls: Type[MonitorConfig],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> MonitorConfig:
        """
        Override the base class `create` to validate database include/exclude
        and derive the `execution_trigger` dict field
        """
        MonitorConfig.database_include_exclude_list_is_valid(data)
        MonitorConfig.derive_execution_trigger_dict(data)
        return super().create(db=db, data=data, check_name=check_name)

    @staticmethod
    def derive_execution_trigger_dict(data: Dict[str, Any]) -> None:
        """
        Determines the execution trigger (cron) dict based on the
        corresponding schema properties provided in the `data` dict.

        The `data` dict is updated in place with the execution trigger dict
        placed in the `monitor_execution_trigger` key, if applicable, and with the
        `execution_frequency` and `execution_start_date` keys removed.

        The `execution_start_date` is inferred as the basis for the day and time
        for repeated monitor execution, and the frequency of execution is based
        on the `execution_frequency` field.

        For example, an `execution_start_date` of "2024-05-14 12:00:00+00:00":
        - with an `execution_frequency` of "daily", it will result in daily
        execution at 12:00:00+00:00;
        - with an `execution_frequency` of "weekly", it will result in weekly
        execution at 12:00:00+00:00 on every Tuesday, since 2024-05-14 is
        a Tuesday.
        - with an `execution_frequency` of "monthly", it will result in monthly
        execution at 12:00:00+00:00 on the 14th day of every month.
        - with an `execution_frequency` of "quarterly", it will result in quarterly
        execution at 12:00:00+00:00 on the 14th day of the first month of each quarter.
        - with an `execution_frequency` of "yearly", it will result in yearly
        execution at 12:00:00+00:00 on May 14th of each year.

        See https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html
        for more information about the cron trigger parameters.
        """
        execution_frequency = data.pop("execution_frequency", None)
        execution_start_date = data.pop("execution_start_date", None)
        if execution_frequency == MonitorFrequency.NOT_SCHEDULED:
            data["monitor_execution_trigger"] = None
            return
        if execution_frequency and execution_start_date:
            cron_trigger_dict = {}
            cron_trigger_dict["start_date"] = execution_start_date
            cron_trigger_dict["timezone"] = execution_start_date.tzinfo
            cron_trigger_dict["hour"] = execution_start_date.hour
            cron_trigger_dict["minute"] = execution_start_date.minute
            cron_trigger_dict["second"] = execution_start_date.second
            if execution_frequency == MonitorFrequency.WEEKLY:
                cron_trigger_dict["day_of_week"] = execution_start_date.weekday()
            if execution_frequency == MonitorFrequency.MONTHLY:
                cron_trigger_dict["day"] = execution_start_date.day
            if execution_frequency == MonitorFrequency.QUARTERLY:
                cron_trigger_dict["day"] = execution_start_date.day
                # Calculate which month of the quarter (0-2) this is
                month_of_quarter = (execution_start_date.month - 1) % 3
                # Set to run in the same month of each quarter (1, 4, 7, 10 for first month)
                cron_trigger_dict["month"] = (
                    f"{1 + month_of_quarter},{4 + month_of_quarter},{7 + month_of_quarter},{10 + month_of_quarter}"
                )
            if execution_frequency == MonitorFrequency.YEARLY:
                cron_trigger_dict["day"] = execution_start_date.day
                cron_trigger_dict["month"] = execution_start_date.month
            data["monitor_execution_trigger"] = cron_trigger_dict


class StagedResource(Base):
    """
    Base DB model that represents a staged resource, fields common to all types of staged resources
    """

    name = Column(String, nullable=True)
    urn = Column(String, index=True, unique=True, nullable=False)
    resource_type = Column(String, index=True, nullable=True)
    description = Column(String, nullable=True)
    monitor_config_id = Column(
        String,
        index=True,  # indexed because we frequently need to slice by monitor config ID
        nullable=True,
    )  # just a "soft" pointer, for now TODO: make this a FK

    # for now, this is just used for web monitor resources.
    system_id = Column(
        String,
        ForeignKey(System.id_field_path),
        nullable=True,
        index=True,
    )

    # TODO: we should be able to enable the below relationship, but it
    # confuses different functionality since 'system' means different
    # things depending on whether the resource is a datastore or web monitor
    # staged resource
    #    system = relationship(System)

    # the Compass vendor ID associated with the StagedResource.
    # only used for web monitor resources
    vendor_id = Column(
        String,
        nullable=True,
        index=True,  # indexed because we frequently need to slice by vendor ID
    )

    source_modified = Column(
        DateTime(timezone=True),
        nullable=True,
    )  # when the resource was modified in the datasource
    classifications = Column(
        ARRAY(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    user_assigned_data_categories = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    user_assigned_data_uses = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    user_assigned_system_id = Column(
        String,
        ForeignKey(System.id_field_path),
        nullable=True,
        index=True,
    )

    # pointers to child and parent URNs
    children = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    parent = Column(String, nullable=True)

    # diff-related fields
    diff_status = Column(String, nullable=True)
    child_diff_statuses = Column(
        MutableDict.as_mutable(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )

    # placeholder for additional attributes
    meta = Column(
        MutableDict.as_mutable(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )

    data_uses = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )

    @classmethod
    def get_urn(cls, db: Session, urn: str) -> Optional[StagedResource]:
        """Utility to retrieve the staged resource with the given URN"""
        return cls.get_by(db=db, field="urn", value=urn)

    @classmethod
    def get_urn_list(cls, db: Session, urns: Iterable[str]) -> Iterable[StagedResource]:
        """
        Utility to retrieve all staged resources with the given URNs
        """
        results = db.execute(select(StagedResource).where(StagedResource.urn.in_(urns)))
        return results.scalars().all()

    @classmethod
    async def get_urn_async(
        cls, db: AsyncSession, urn: str
    ) -> Optional[StagedResource]:
        """
        Utility to retrieve the staged resource with the given URN using an async session
        """
        results = await db.execute(
            select(StagedResource).where(StagedResource.urn == urn)
        )
        return results.scalars().first()

    @classmethod
    async def get_urn_list_async(
        cls, db: AsyncSession, urns: List[str]
    ) -> Optional[List[StagedResource]]:
        """
        Utility to retrieve the staged resource with the given URN using an async session
        """
        results = await db.execute(
            select(StagedResource).where(StagedResource.urn.in_(urns))
        )
        return results.scalars().all()

    def add_child_diff_status(self, diff_status: DiffStatus) -> None:
        """Increments the specified child diff status"""
        self.child_diff_statuses[diff_status.value] = (
            self.child_diff_statuses.get(diff_status.value, 0) + 1
        )

    def mark_as_addition(
        self,
        db: Session,
        parent_resource_urns: Iterable[str] = [],
    ) -> None:
        """
        Marks the resource as an addition and the child diff status of
        the given parent resource URNs accordingly
        """
        self.diff_status = DiffStatus.ADDITION.value
        for parent_resource_urn in parent_resource_urns:
            parent_resource: Optional[StagedResource] = StagedResource.get_urn(
                db, parent_resource_urn
            )
            if parent_resource:
                parent_resource.add_child_diff_status(DiffStatus.ADDITION)


class MonitorExecution(Base):
    """
    Monitor execution record used for data detection and discovery.

    Each monitor execution references `MonitorConfig`, which provide it with underlying
    configuration details used in connecting to the external data store.
    """

    monitor_config_key = Column(
        String,
        ForeignKey(MonitorConfig.key),
        nullable=False,
        index=True,
    )
    status = Column(String, nullable=True)
    started = Column(
        DateTime(timezone=True),
        nullable=True,
        server_default=func.now(),
    )
    completed = Column(DateTime(timezone=True), nullable=True)
    classification_instances = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        default=list,
    )
    # stores additional information from monitor execution failures as an array of strings,
    # e.g. error messages, stack traces, etc.
    messages = Column(
        ARRAY(String),
        nullable=False,
        default=list,
    )
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )


def fetch_staged_resources_by_type_query(
    resource_type: str,
    monitor_config_ids: Optional[List[str]] = None,
    show_hidden: bool = False,
) -> Query[StagedResource]:
    """
    Fetches staged resources by type and monitor config ID. Optionally filters out muted staged resources ("hidden").
    """
    logger.info(
        f"Fetching staged resources of type {resource_type}, show_hidden={show_hidden}, monitor_config_ids={monitor_config_ids}"
    )
    query = select(StagedResource).where(StagedResource.resource_type == resource_type)

    if monitor_config_ids:
        query = query.filter(StagedResource.monitor_config_id.in_(monitor_config_ids))
    if not show_hidden:
        from sqlalchemy import or_

        query = query.filter(
            or_(
                StagedResource.diff_status != DiffStatus.MUTED.value,
                StagedResource.diff_status.is_(None),
            )
        )

    return query



from enum import Enum
from os.path import dirname, join
from typing import Dict

import yaml
from pydantic import BaseModel

from fides.api.util.text import to_snake_case
from fides.config.helpers import load_file

LANGUAGES_YAML_FILE_PATH = join(
    dirname(__file__),
    "../../data",
    "language",
    "languages.yml",
)


class Language(BaseModel):
    id: str
    name: str


def _load_supported_languages() -> Dict[str, Language]:
    """Loads language dict based on yml file on disk"""
    with open(load_file([LANGUAGES_YAML_FILE_PATH]), "r", encoding="utf-8") as file:
        _languages = yaml.safe_load(file).get("languages", [])
        language_dict = {}
        for language in _languages:
            language_dict[language["id"]] = Language.model_validate(language)
        return language_dict


supported_languages_by_id: Dict[str, Language] = (
    _load_supported_languages()
)  # should only be accessed for read-only access


# dynamically create an enum based on definitions loaded from YAML
SupportedLanguage: Enum = Enum(  # type: ignore[misc]
    "SupportedLanguage",
    {
        to_snake_case(language.name): language.id
        for language in supported_languages_by_id.values()
    },
)



"""
Reusable command-line arguments and options.
"""

from typing import Callable

import rich_click as click
from fideslang import model_list

# Note - the type: ignore comments throughout this file are a workaround
# for a known issue with the `click` library:
# https://github.com/pallets/click/issues/2558
# If/when that issue is resolved, they can be removed.


def diff_flag(command: Callable) -> Callable:
    """Print any diffs between the local & server objects"""
    command = click.option(
        "--diff",
        is_flag=True,
        help="Print any diffs between the local & server objects",
    )(
        command
    )  # type: ignore
    return command


def coverage_threshold_option(command: Callable) -> Callable:
    """An option decorator that sets a required coverage percentage."""
    command = click.option(
        "--coverage-threshold",
        "-c",
        type=click.IntRange(0, 100),
        default=100,
        help="Set the coverage percentage for a passing scan.",
    )(
        command
    )  # type: ignore
    return command


def resource_type_argument(command: Callable) -> Callable:
    "Add the resource_type argument."
    command = click.argument(
        "resource_type",
        type=click.Choice(model_list, case_sensitive=False),
    )(
        command
    )  # type: ignore
    return command


def resource_type_option(command: Callable) -> Callable:
    "Add the resource_type option."
    command = click.option(
        "--resource-type",
        default="",
        help=f"Choose from {str(model_list)}",
    )(
        command
    )  # type: ignore
    return command


def fides_key_argument(command: Callable) -> Callable:
    "Add the fides_key argument."
    command = click.argument(
        "fides_key",
        type=str,
    )(
        command
    )  # type: ignore
    return command


def fides_key_option(command: Callable) -> Callable:
    "Add the fides_key option."
    command = click.option(
        "-k",
        "--fides-key",
        default="",
    )(
        command
    )  # type: ignore
    return command


def manifests_dir_argument(command: Callable) -> Callable:
    "Add the manifests_dir argument."
    command = click.argument(
        "manifests_dir", type=click.Path(exists=True), default=".fides/"
    )(
        command
    )  # type: ignore
    return command


def dry_flag(command: Callable) -> Callable:
    "Add a flag that prevents side-effects."
    command = click.option(
        "--dry", is_flag=True, help="Do not upload results to the Fides webserver."
    )(
        command
    )  # type: ignore
    return command


def yes_flag(command: Callable) -> Callable:
    "Add a flag that assumes yes."
    command = click.option(
        "--yes",
        "-y",
        is_flag=True,
        help="Automatically responds `yes` to any prompts.",
    )(
        command
    )  # type: ignore
    return command


def verbose_flag(command: Callable) -> Callable:
    "Turns on verbose output."
    command = click.option(
        "--verbose",
        "-v",
        is_flag=True,
        help="Enable verbose output.",
    )(
        command
    )  # type: ignore
    return command


def include_null_flag(command: Callable) -> Callable:
    "Include null attributes in resource output."
    command = click.option(
        "--include-null",
        is_flag=True,
        help="Include null attributes.",
    )(
        command
    )  # type: ignore
    return command


def organization_fides_key_option(command: Callable) -> Callable:
    "Add the organization_fides_key option."
    command = click.option(
        "--org-key",
        "-k",
        default="default_organization",
        show_default=True,
        help="The `organization_fides_key` of the `Organization` you want to specify.",
    )(
        command
    )  # type: ignore
    return command


def output_directory_option(command: Callable) -> Callable:
    "Add the output directory option"
    command = click.option(
        "--output-dir",
        "-d",
        default=".fides/",
        show_default=True,
        help="The output directory for the data map to be exported to.",
    )(
        command
    )  # type: ignore
    return command


def credentials_id_option(command: Callable) -> Callable:
    "Use credentials defined within fides config."
    command = click.option(
        "--credentials-id",
        type=str,
        help="Use credentials keys defined within Fides config.",
    )(
        command
    )  # type: ignore
    return command


def connection_string_option(command: Callable) -> Callable:
    "Use connection string option to connect to a database"
    command = click.option(
        "--connection-string",
        type=str,
        help="Use the connection string option to connect to a database.",
    )(
        command
    )  # type: ignore
    return command


def okta_org_url_option(command: Callable) -> Callable:
    "Use org url option to connect to okta. Requires options --org-url and --token"
    command = click.option(
        "--org-url",
        type=str,
        help="Connect to Okta using an 'Org URL'. _Requires options `--org-url` & `--token`._",
    )(
        command
    )  # type: ignore
    return command


def okta_token_option(command: Callable) -> Callable:
    "Use token option to connect to okta. Requires options --org-url and --token"
    command = click.option(
        "--token",
        type=str,
        help="Connect to Okta using a token. _Requires options `--org-url` and `--token`._",
    )(
        command
    )  # type: ignore
    return command


def aws_access_key_id_option(command: Callable) -> Callable:
    "Use access key id option to connect to aws. Requires options --access_key_id, access_key and --region"
    command = click.option(
        "--access_key_id",
        type=str,
        help="Connect to AWS using an `Access Key ID`. _Requires options `--access_key_id`, `--secret_access_key` & `--region`._",
    )(
        command
    )  # type: ignore
    return command


def aws_secret_access_key_option(command: Callable) -> Callable:
    "Use access key option to connect to aws. Requires options --access_key_id, --secret_access_key and --region"
    command = click.option(
        "--secret_access_key",
        type=str,
        help="Connect to AWS using an `Access Key`. _Requires options `--access_key_id`, `--secret_access_key` & `--region`._",
    )(
        command
    )  # type: ignore
    return command


def aws_session_token_option(command: Callable) -> Callable:
    "Use a session token with temporary access keys to connect to aws. Requires options --access_key_id, --secret_access_key and --region"
    command = click.option(
        "--session_token",
        type=str,
        default="",
        help="Connect to AWS using a temporary `Access Key`. _Requires options `--access_key_id`, `--secret_access_key`, `--session_token`, & `--region`._",
    )(
        command
    )  # type: ignore
    return command


def aws_region_option(command: Callable) -> Callable:
    "Use region option to connect to aws. Requires options --access_key_id, access_key and --region"
    command = click.option(
        "--region",
        type=str,
        help="Connect to AWS using a specific `Region`. _Requires options `--access_key_id`, `--secret_access_key` & `--region`._",
    )(
        command
    )  # type: ignore
    return command


def prompt_username(ctx: click.Context, param: str, value: str) -> str:
    """Check the config for a compatible username. If unavailable, prompt the user to provide one."""
    config_value = ctx.obj["CONFIG"].user.username

    if value:
        return value

    if config_value:
        click.echo("> Username found in configuration.")
        return config_value

    value = click.prompt(text="Username")  # pragma: no cover
    return value


def prompt_password(ctx: click.Context, param: str, value: str) -> str:
    """Check the config for a compatible password. If unavailable, prompt the user to provide one."""
    config_value = ctx.obj["CONFIG"].user.password

    if value:
        return value

    if config_value:
        click.echo("> Password found in configuration.")
        return config_value

    value = click.prompt(text="Password", hide_input=True)  # pragma: no cover
    return value


def username_option(command: Callable) -> Callable:
    command = click.option(
        "-u",
        "--username",
        help="If not provided, will be pulled from the config file or prompted for.",
        default="",
        callback=prompt_username,
    )(
        command
    )  # type: ignore
    return command


def password_option(command: Callable) -> Callable:
    command = click.option(
        "-p",
        "--password",
        help="If not provided, will be pulled from the config file or prompted for.",
        default="",
        callback=prompt_password,
    )(
        command
    )  # type: ignore
    return command


def first_name_option(command: Callable) -> Callable:
    command = click.option(
        "-f",
        "--first-name",
        default="",
    )(
        command
    )  # type: ignore
    return command


def last_name_option(command: Callable) -> Callable:
    command = click.option(
        "-l",
        "--last-name",
        default="",
    )(
        command
    )  # type: ignore
    return command


def username_argument(command: Callable) -> Callable:
    command = click.argument("username", type=str)(command)  # type: ignore
    return command


def password_argument(command: Callable) -> Callable:
    command = click.argument("password", type=str)(command)  # type: ignore
    return command


def email_address_argument(command: Callable) -> Callable:
    command = click.argument("email_address", type=str)(command)  # type: ignore
    return command



"""add field for excluding databases on MonitorConfig model

Revision ID: f712aa9429f4
Revises: 31493e48c1d8
Create Date: 2024-07-11 18:00:14.221036

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f712aa9429f4"
down_revision = "31493e48c1d8"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "monitorconfig",
        sa.Column(
            "excluded_databases",
            sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
    )


def downgrade():
    op.drop_column("monitorconfig", "excluded_databases")
